from axon.config import default_rpc_endpoint, default_service_config, default_service_depth
from axon.ServiceNode import ServiceNode, transport_layers
from axon.utils import overwrite

from threading import Thread

registered_ServiceNodes = {}
TLSNs = {}

def get_TLSN(configuration):
	global TLSNs
	configuration = overwrite(default_service_config, configuration)

	top_service_node = None
	tl_id = id(configuration['tl'])

	# record each transport layer we see, and create a top-level ServiceNode for each of them
	# if default_service_config['endpoint_prefix'] in configuration['tl'].rpcs:
	if tl_id in TLSNs:
		top_service_node = TLSNs[tl_id]

	else:
		top_service_node = ServiceNode({}, '', **configuration)
		top_service_node.add_child(default_rpc_endpoint, object())
		TLSNs[tl_id] = top_service_node

	return top_service_node

def service(subject, name, depth=default_service_depth, **configuration):
	global TLSNs

	s = ServiceNode(subject, name, depth=depth, **configuration)
	registered_ServiceNodes[name] = s

	top_service_node = get_TLSN(configuration)
	top_service_node.add_child(name, subject, **configuration)

	return s

# the RPC decorator adds the associated function to the RPC_node ServiceNode
def rpc(**configuration):
	global TLSNs

	top_service_node = get_TLSN(configuration)
	RPC_node = top_service_node.children[default_rpc_endpoint]

	def add_to_RPC_node(fn):
		RPC_node.add_child(fn.__name__, fn, **configuration)
		return fn

	return add_to_RPC_node

def init(tl=default_service_config['tl']):

	transport_layers.add(tl)

	for i in range(len(transport_layers)-1):
		t = transport_layers.pop()
		tl_thread = Thread(target=t.run, daemon=True)
		tl_thread.start()

	last_tl = transport_layers.pop()
	last_tl.run()


from concurrent.futures import Future, Executor

class InlineExecutor(Executor):

	def submit(self, target, *args, **kwargs):
		result_future = Future()

		kwargs['in_parallel'] = False

		result_future.set_result(target(*args, **kwargs))
		return result_future



import asyncio
import urllib3
import concurrent.futures as futures
import threading
import inspect

from abc import ABC, abstractmethod

from axon.serializers import serialize, deserialize

req_executor = futures.ThreadPoolExecutor(max_workers=100)
http = urllib3.PoolManager(maxsize=100)

# this function checks if an error flag has been set and raises the corresponding error if it has
def error_handler(result_str):
	err_code, result_str = result_str.split('|', 1)

	if (err_code == '1'):
		# an error occured in worker, raise it
		(error_info, error) = deserialize(result_str)

		print('the following error occured in worker:')
		print(error_info)
		raise(error)

	else:
		return result_str


class AsyncResultHandle():

	def __init__(self, fn, *args, **kwargs):
		self.fn = fn
		self.args = args
		self.kwargs = kwargs

	def __await__(self):
		loop = asyncio.get_event_loop()
		return loop.run_in_executor(req_executor, self.fn, *self.args, **self.kwargs).__await__()

	def join(self):
		f = req_executor.submit(self.fn, *self.args, **self.kwargs)
		return f.result()

class AbstractTransportClient(ABC):

	@abstractmethod
	def get_config(self):
		pass

	@abstractmethod
	def call_rpc(self, url, args, kwargs):
		pass



termination_str = 'End of message'

def send_in_chunks(socket, input_str, chunk_size=100_000):

	for i in range(len(input_str)//chunk_size):
		chunk = input_str[ chunk_size*i : chunk_size*(i+1) ]
		socket.send(chunk)

	socket.send(input_str[chunk_size*(len(input_str)//chunk_size):])
	socket.send(termination_str)

def recv_chunks(socket):

	result_str = ''
	while True:
		chunk = socket.recv()

		if (chunk == termination_str):
			break

		else:
			result_str += chunk

	return result_str


# this file starts a thread with an axon worker in it, for the purposes of testing client code

import threading
import axon
import time
import psutil

# starts a worker thread, fixtures in some tests will add services and RPCs to run tests off this worker instance
# without this check, using the multiprocessing executor will result in this file being run more than once in other processes, and so we must check if we're in the main process before calling init
if (psutil.Process().name() == 'pytest'):

	worker_thread = threading.Thread(target=axon.worker.init, daemon=True, name='axon/tests/__init__.py')
	worker_thread.start()
	time.sleep(1)


from sys import path
path.append('..')

from axon.serializers import serialize, deserialize
from axon.transport_client import AbstractTransportClient, http, error_handler
from axon.HTTP_transport import config

class HTTPTransportClient(AbstractTransportClient):

	def __init__(self):
		pass

	def get_config(self):
		return config

	def call_rpc(self, url, args, kwargs):

		resp = http.request('POST', url, fields={'msg': serialize((args, kwargs))})
		result_str = error_handler(resp.data.decode())
		return deserialize(result_str)


import axon
import asyncio
import time
import json
import pytest

from axon.tests.client_test import fix_basic_service

default_service_depth = axon.config.default_service_depth
port = axon.config.transport.config.port
url_scheme = axon.config.url_scheme
TransportClient = type(axon.config.default_client_tl)

# the endpoint that our service will be located at
endpoint = 'test_endpoint_prefix'
# the name of our service
service_name = 'basic_service'

test_service_depth = 3

# this test manulually creates a stub that points to a service endpoint. Note that each endpoint is suffixed with /__call__ since RPC configs are stored on the __call__ attribute
@pytest.mark.asyncio
async def test_GenericStub(fix_basic_service):
	print('test_GenericStub')

	# tests that ServiceStubs, including child stubs, are instantiated properly and that their RPCs work
	full_endpoint = f'{endpoint}/{service_name}'
	for i in range(default_service_depth, 0, -1):
		print('testing GenericStub')

		url=f'{url_scheme}://localhost:{port}/{full_endpoint}/test_fn/__call__'
		stub = axon.stubs.GenericStub(url=url, tl=TransportClient())
		
		await stub()

		# next iteration, the stub will point at the child service
		full_endpoint = full_endpoint + '/child'

# this test creates a metastub to a test service and calls methods recursively to check each child object. Also checks inheritance from a BaseClass
@pytest.mark.asyncio
async def test_MetaServiceStub():
	print('test_MetaServiceStub')

	class BaseClass():
		def __init__(self):
			pass

	url = f'{url_scheme}://localhost:{port}/{endpoint}/{service_name}'
	worker = axon.client.get_stub(url, top_stub_type=BaseClass)

	# Tests that the stub is inherited from BaseClass, as specified by kwarg top_stub_type
	if isinstance(worker, BaseClass):
		print('Inheritance from BaseClass confirmed')
	else:
		raise BaseException('Stub is not inheritance from BaseClass')

	# tests that child stubs are instantiated properly and that their RPCs work
	for i in range(test_service_depth, 0, -1):
		await worker.test_fn()
		await worker()

		if isinstance(worker.test_fn, axon.stubs.GenericStub):
			print('RPC inheritance from axon.stubs.GenericStub confirmed')
		else:
			raise BaseException('RPC stub is not inheritance from axon.stubs.GenericStub')

		if isinstance(worker, axon.stubs.GenericStub):
			print('Callable stub inheritance from axon.stubs.GenericStub confirmed')
		else:
			raise BaseException('Callable stub is not inheritance from axon.stubs.GenericStub')

		if (i != 1):
			# if this is the last iteration, worker won't have a child and this line will raise an attribute error
			worker = worker.child

# this test creates a metastub to a test service that inherits from SyncStubs and calls methods recursively to check that each child function is a sync stub
@pytest.mark.asyncio
async def test_SyncStub():
	print('test_SyncStub')

	url = f'{url_scheme}://localhost:{port}/{endpoint}/{service_name}'
	worker = axon.client.get_stub(url, stub_type=axon.stubs.SyncStub)

	# tests that child stubs are instantiated properly and that their RPCs work
	for i in range(test_service_depth, 0, -1):
		worker.test_fn()
		worker()

		# tests that stub is inherited from SyncStub
		if isinstance(worker.test_fn, axon.stubs.SyncStub):
			print('Inheritance from axon.stubs.SyncStub confirmed')
		else:
			raise BaseException('Stub is not inheritance from axon.stubs.SyncStub')

		if (i != 1):
			# if this is the last iteration, worker won't have a child and this line will raise an attribute error
			worker = worker.child


import pickle
import codecs

# pickle operates on bytes, but http operates on strings, so we've gotta convert pickles to and from a string
def serialize(obj):
	pickled = pickle.dumps(obj)
	return codecs.encode(pickled, "base64").decode()

# pickle operates on bytes, but http operates on strings, so we've gotta convert pickles to and from a string
def deserialize(obj_str):
	# print(obj_str)
	obj_bytes = codecs.decode(obj_str.encode(), "base64")
	# print(pickle.loads(obj_bytes))
	return pickle.loads(obj_bytes)


import axon
import threading
import time
import pytest
import random
import cloudpickle

from concurrent.futures import ThreadPoolExecutor

refl_http_port=8081
tpe = ThreadPoolExecutor(max_workers=10)

@pytest.fixture(scope='package')
def refl_thread():

	reflector_thread = threading.Thread(target=axon.reflector.run, kwargs={'http_port':refl_http_port}, daemon=True)
	reflector_thread.start()
	time.sleep(1)

class DummyClass():

	def __init__(self):
		pass

	def print_str(self, str):
		print('===========================')
		print(f"here's the string: {str}")
		print('===========================')

		return 'all done!'

@pytest.fixture(scope='package')
def test_service(refl_thread):

	itlw = axon.reflector.ITLW(url='localhost', name='test_service')

	axon.worker.service(DummyClass(), 'test_service', tl=itlw, executor=tpe)

	worker_thread = threading.Thread(target=itlw.run, daemon=True)
	worker_thread.start()
	time.sleep(1)

@pytest.fixture(scope='package')
def echo_worker(refl_thread):

	itlw = axon.reflector.ITLW(url='localhost', name='echo_worker')

	@axon.worker.rpc(tl=itlw, executor=tpe)
	def echo(msg):
		return msg

	worker_thread = threading.Thread(target=itlw.run, daemon=True)
	worker_thread.start()
	time.sleep(1)

def test_basic_operation(echo_worker, test_service):

	stub = axon.client.get_stub(f'{axon.config.url_scheme}://localhost:{refl_http_port}/reflected_services')

	test_msg = 'this is a message sent from client to reflector, then to worker, and then back again'
	response = stub.echo_worker.rpc.echo(test_msg).join()
	assert(response == test_msg)

	response = stub.test_service.test_service.print_str(test_msg).join()
	assert(response == 'all done!')

def test_chunking(echo_worker):
	stub = axon.client.get_stub(f'{axon.config.url_scheme}://localhost:{refl_http_port}/reflected_services')

	# creates a large string to test if the chunking feature is working, since the data should be larger than the max message size in SocketIO
	msg_size = 1_000_000
	msg = ''.join([str(random.randint(0,9)) for i in range(msg_size)])

	response = stub.echo_worker.rpc.echo(msg).join()
	
	assert response == msg

@pytest.fixture(scope='package')
def host_worker(refl_thread):

	itlw = axon.reflector.ITLW(url='localhost', name='host_worker')

	@axon.worker.rpc(tl=itlw, executor=tpe)
	def host(service_str, name):
		service = cloudpickle.loads(service_str)
		axon.worker.service(service, name, tl=itlw)
		itlw.update_profile()
		return 'success'

	worker_thread = threading.Thread(target=itlw.run, daemon=True)
	worker_thread.start()
	time.sleep(1)

def test_hosting(host_worker):
	stub = axon.client.get_stub(f'{axon.config.url_scheme}://localhost:{refl_http_port}/reflected_services')

	t = DummyClass()
	r = stub.host_worker.rpc.host(cloudpickle.dumps(t), 'hosted_t').join()
	assert(r == 'success')

	stub = axon.client.get_stub(f'{axon.config.url_scheme}://localhost:{refl_http_port}/reflected_services/host_worker/hosted_t')
	resp = stub.print_str('This comes from the host!').join()
	assert(resp == 'all done!')

def test_disconnect(refl_thread):

	itlw = axon.reflector.ITLW(url='localhost', name='disconnect_worker')

	@axon.worker.rpc(tl=itlw, executor=tpe)
	def delay(i):
		time.sleep(i)

	worker_thread = threading.Thread(target=itlw.run, daemon=True)
	worker_thread.start()
	time.sleep(1)

	stub = axon.client.get_stub(f'{axon.config.url_scheme}://localhost:{refl_http_port}/reflected_services')

	def disconnect_fn():
		time.sleep(0.5)
		itlw.sio.disconnect()

	disconnect_thread = threading.Thread(target=disconnect_fn, daemon=True)
	disconnect_thread.start()

	with pytest.raises(BaseException) as err:
		stub.disconnect_worker.rpc.delay(1).join()

	assert str(err.value) == 'WorkerDisconnect'


import sys
sys.path.append('..')

import axon
from concurrent.futures import ThreadPoolExecutor

itlw = axon.reflector.ITLW(url='143.198.32.69', name='test_worker')
tpe = ThreadPoolExecutor(max_workers=10)

@axon.worker.rpc(tl=itlw)
def echo(msg):
	raise BaseException('this is an error!')
	return msg

axon.worker.init()


from axon import axon
from types import SimpleNamespace

import pytest
import asyncio
import time

url_scheme = axon.config.url_scheme
TransportClient = type(axon.config.default_client_tl)
tl_config = axon.config.transport.config

@pytest.fixture(scope='package')
def fix_basic_rpc():
	@axon.worker.rpc()
	def basic_rpc(prefix, suffix='basic test failed'):
		return prefix+suffix

@pytest.mark.tl
def test_tl_client(fix_basic_rpc):
	print('test_tl_client')

	rpc_name = 'basic_rpc'
	url = f'{url_scheme}://localhost:{tl_config.port}/{axon.config.default_rpc_endpoint}/{rpc_name}/__call__'
	tl = TransportClient()

	assert('test passed!' == tl.call_rpc(url, ('test ', ), {'suffix':'passed!', }))

	url = f'{url_scheme}://localhost:{tl_config.port}/{axon.config.default_service_config["endpoint_prefix"]}'
	profile = tl.call_rpc(url, (), {})

	assert('rpc' in profile)
	assert('basic_rpc' in profile['rpc'])

url_shorthand_test_cases = [('http://localhost:50/endpoint', 'http://localhost:50/endpoint'), ('http://localhost:50/endpoint', 'http://localhost/endpoint'), ('http://localhost:50/endpoint', 'localhost:50/endpoint'), ('http://localhost:50/endpoint', 'localhost/endpoint')]
url_shorthand_test_cases += [('http://192.168.2.0:50/endpoint', 'http://192.168.2.0:50/endpoint'), ('http://192.168.2.0:50/endpoint', 'http://192.168.2.0/endpoint'), ('http://192.168.2.0:50/endpoint', '192.168.2.0:50/endpoint'), ('http://192.168.2.0:50/endpoint', '192.168.2.0/endpoint')]
url_shorthand_test_cases += [('http://www.google.com:50/endpoint', 'http://www.google.com:50/endpoint'), ('http://www.google.com:50/endpoint', 'http://www.google.com/endpoint'), ('http://www.google.com:50/endpoint', 'www.google.com:50/endpoint'), ('http://www.google.com:50/endpoint', 'www.google.com/endpoint')]

@pytest.mark.parametrize("expected, shorthand", url_shorthand_test_cases)
def test_url_shorthand(expected, shorthand):
	print('test_url_shorthand')

	config = SimpleNamespace(
		port=50,
		scheme='http'
	)

	assert axon.stubs.add_url_defaults(shorthand, config) == expected

class RecursiveCallable():
	def __init__(self, depth=1):
		self.depth = depth

		if (depth>1):
			self.child = RecursiveCallable(depth=self.depth-1)
		else:
			self.child = None

	def test_fn(self):
		print(f'test_fn called at depth {self.depth}')

	def __call__(self):
		print(f'__call__ called at depth {self.depth}')

@pytest.fixture(scope='package')
def fix_basic_service():
	# the endpoint that our service will be located at
	endpoint = '/test_endpoint_prefix'

	# defines an instance of RecursiveCallable and creates a service node out of it
	test_service_depth = 3
	t_basic = RecursiveCallable(depth=test_service_depth)

	basic_service = axon.worker.service(t_basic, 'basic_service', depth=test_service_depth, endpoint_prefix=endpoint)

@pytest.mark.asyncio
async def test_TopLevelServiceNode(fix_basic_service):
	print('test_TopLevelServiceNode')

	w = axon.client.get_stub(f'{url_scheme}://localhost:{tl_config.port}')

	print(await w.rpc.basic_rpc('basic test ', suffix='passed'))

	# tests that the ServiceStub is set up right
	stub = w.basic_service

	# tests that child stubs are instantiated properly and that their RPCs work
	for i in range(axon.config.default_service_depth, 0, -1):
		await w.basic_service()
		w.basic_service().join()

		await w.basic_service.test_fn()
		w.basic_service.test_fn().join()

		if isinstance(w.basic_service, axon.stubs.GenericStub):
			print('Callable Stub inheritance from axon.stubs.GenericStub confirmed')
		else:
			raise BaseException('Callable Stub stub is not inheritance from axon.stubs.GenericStub')

		if isinstance(w.basic_service.test_fn, axon.stubs.GenericStub):
			print('RPC inheritance from axon.stubs.GenericStub confirmed')
		else:
			raise BaseException('RPC stub is not inheritance from axon.stubs.GenericStub')

		if (i != 1):
			# if this is the last iteration, worker won't have a child and this line will raise an attribute error
			w.basic_service = w.basic_service.child


from axon.serializers import deserialize, serialize

from flask import Flask
from flask import request as route_req
from concurrent.futures import ProcessPoolExecutor as PPE
from threading import Thread, Lock
from abc import ABC, abstractmethod

import sys
import logging
import random
import string
import cloudpickle
import traceback
import inspect
import pickle
import asyncio
import time

loop, event_loop_thread = None, None
def start_event_loop_thread():
	global loop, event_loop_thread
	loop = asyncio.new_event_loop()
	asyncio.set_event_loop(loop)
	event_loop_thread = Thread(target=loop.run_forever, daemon=True)
	event_loop_thread.start()

inline_lock = Lock()
def invoke_RPC(target_fn, param_str, in_parallel=True):
	global loop, event_loop_thread

	if isinstance(target_fn, bytes) or isinstance(target_fn, str):
		target_fn = pickle.loads(target_fn)

	args, kwargs = deserialize(param_str)

	result = None

	if not in_parallel:
		with inline_lock:
			result = target_fn(*args, **kwargs)
	else:
		result = target_fn(*args, **kwargs)

	if inspect.iscoroutine(result):
		if (event_loop_thread == None) or not (event_loop_thread.is_alive()):
			start_event_loop_thread()
			
		result = asyncio.run_coroutine_threadsafe(result, loop).result()

	return serialize(result)

class AbstractTransportWorker(ABC):

	@abstractmethod
	def run(self):
		pass

	def register_RPC(self, fn, endpoint, executor):

		if isinstance(executor, PPE):
			fn = cloudpickle.dumps(fn)

		self.rpcs[endpoint] = (fn, executor)

	def deregister_RPC(self, endpoint):

		if endpoint in self.rpcs:
			del self.rpcs[endpoint]
		else:
			raise BaseException(f'No RPC registered at endpoint: {endpoint}')


import sys
sys.path.append('..')

import axon
import time
import threading
import socketio
import logging
import psutil
import random

from concurrent.futures import Future
from threading import Timer
from flask import Flask

from axon.serializers import serialize, deserialize
from axon.transport_client import req_executor, error_handler, AsyncResultHandle
from axon.config import transport, default_service_config
from axon.chunking import send_in_chunks, recv_chunks
from axon.HTTP_transport.config import port as default_http_port
from axon.utils import get_ID_generator
from axon.reflector import config as refl_config
from axon.transport_client import AbstractTransportClient

sio = socketio.Server(async_mode='threading')
reflector_node = None
client_sid_map = {}

logger = None
def init_logger():
	global logger

	logger = logging.getLogger(__name__)
	logger.setLevel(logging.DEBUG)

	c_handler = logging.StreamHandler()
	f_handler = logging.FileHandler(refl_config.log_file)

	log_format = logging.Formatter('%(asctime)s | %(levelname)s | %(message)s')
	c_handler.setFormatter(log_format)
	f_handler.setFormatter(log_format)

	logger.addHandler(c_handler)
	logger.addHandler(f_handler)

# this class extends the client and encapsulates the connection with a worker
class ITL_Client(AbstractTransportClient):

	def __init__(self, sio, sid, name):
		self.sio = sio
		self.sid = sid
		self.name = name
		self.pending_reqs = {}
		self.chunk_buffers = {}
		self.call_ID_gen = get_ID_generator()

	def get_config(self):
		# the ITL client sends requests through an already established socket connection, so config info like the port number and scheme don't exist
		return None

	def call_rpc(self, url, args, kwargs):

		url_components = url.split('/')
		url_head = '/'.join(url_components[:3])
		endpoint = '/' + '/'.join(url_components[3:])

		call_ID = next(self.call_ID_gen)
		result_future = Future()
		self.pending_reqs[call_ID] = result_future

		logger.debug('RPC call to: %s for: %s call_ID: %s', self.sid, endpoint, call_ID)
		self.sio.emit('rpc_request', to=self.sid, data=f'{call_ID}|{endpoint}|{serialize((args, kwargs))}')
		
		result_str = result_future.result()
		result_str = error_handler(result_str)
		return deserialize(result_str)

	def disconnect_handler(self):

		# send a worker disconnect error back through each pending request
		for call_ID in self.pending_reqs:
			result_str = serialize(('The worker closed connection before responding to RPC', BaseException('WorkerDisconnect')))
			result_str = f'1|{result_str}'
			self.pending_reqs[call_ID].set_result(result_str)

@sio.event
def rpc_result(sid, return_str):
	global client_sid_map

	call_ID, result_str = return_str.split('|', 1)
	logger.debug('RPC response for call_ID: %s', call_ID)
	client = client_sid_map[sid]
	result_future = client.pending_reqs[call_ID]
	result_future.set_result(result_str)

@sio.event
def rpc_result_chunk(sid, res_str):
	global client_sid_map

	chunk_num, num_chunks, call_ID, chunk_str = res_str.split('|', 3)
	logger.debug('RPC response chunk %s for call_ID: %s', chunk_num, call_ID)

	chunk_obj = {
		'chunk_str': chunk_str,
		'chunk_num': int(chunk_num)
	}

	client = client_sid_map[sid]

	if (call_ID in client.chunk_buffers):
		client.chunk_buffers[call_ID].append(chunk_obj)

	else :
		client.chunk_buffers[call_ID] = [chunk_obj]

	if (len(client.chunk_buffers[call_ID]) == int(num_chunks)):

		chunks = client.chunk_buffers[call_ID]
		chunks.sort(key=lambda x: x['chunk_num'])
		chunk_strs = [b['chunk_str'] for b in chunks]
		result_str = ''.join(chunk_strs)

		result_future = client.pending_reqs[call_ID]
		result_future.set_result(result_str)

		logger.debug('recieved all chunks for call_ID: %s', call_ID)
		del client.chunk_buffers[call_ID]

@sio.event
def connect(sid, e):
	logger.debug('New connection from: %s', sid)

@sio.event
def disconnect(sid):
	global reflector_node, client_sid_map

	logger.debug('Worker %s disconnected', sid)
	client = client_sid_map[sid]

	reflector_node.remove_child(client.name)
	client.disconnect_handler()
	del client_sid_map[sid]

@sio.event
def worker_header(sid, name):
	global client_sid_map
	client_sid_map[sid] = ITL_Client(sio, sid, name)
	
@sio.event
def update_profile(sid, profile_str):
	global reflector_node, client_sid_map
	logger.debug(f'update_profile {sid}')

	profile = deserialize(profile_str)

	itl = client_sid_map[sid]
	stub = axon.client.make_ServiceStub('ws://none:0000', itl, profile, stub_type=axon.stubs.SyncStub)
	reflector_node.add_child(itl.name, stub)

def run(endpoint='reflected_services', ws_port=5000, http_port=default_http_port):
	global reflector_node, http_tl, logger

	if logger == None:
		init_logger()

	http_tl = transport.worker(http_port)

	http_thread = threading.Thread(target=http_tl.run, daemon=True)
	http_thread.start()
	time.sleep(0.5)

	# reflector_node = axon.worker.service({}, endpoint, tl=http_tl)
	reflector_node = axon.worker.ServiceNode({}, endpoint, tl=http_tl)

	logger.debug('Reflector start')

	app = Flask(__name__)
	app.wsgi_app = socketio.WSGIApp(sio, app.wsgi_app)
	app.run(host='0.0.0.0', port=ws_port)


from sys import argv as cli_args
from importlib import import_module
from types import SimpleNamespace
from axon.inline_executor import InlineExecutor

TL_name = 'axon.HTTP_transport'

if '--tl' in cli_args:
	i = cli_args.index('--tl')
	TL_name = cli_args[i+1]

transport = import_module(TL_name)
inline_executor = InlineExecutor()
version = "0.2.4"
default_service_depth = 3
default_rpc_endpoint = 'rpc'
NUM_OPEN_REQS = 512

url_scheme = transport.config.scheme
default_client_tl = transport.client()
default_worker_tl = transport.worker()

comms_config = {
	'notice_board_port': 8002,
	'request_timeout': 30,
}

comms_config = SimpleNamespace(**comms_config)

default_service_config = {
	'endpoint_prefix': '',
	'executor': inline_executor,
	'tl': default_worker_tl
}


import pytest
import threading
import asyncio
import time

import axon

from axon.tests.ServiceNode_test import fix_live_transport_worker

url_scheme = axon.config.url_scheme
TransportClient = type(axon.config.default_client_tl)
TransportWorker = type(axon.config.default_service_config['tl'])

@pytest.mark.tl
def test_interface_compliance():

	assert(hasattr(axon.config.transport.config, 'port'))
	assert(hasattr(axon.config.transport.config, 'scheme'))

	assert(isinstance(axon.config.default_client_tl, axon.transport_client.AbstractTransportClient))
	assert(isinstance(axon.config.default_service_config['tl'], axon.transport_worker.AbstractTransportWorker))

@pytest.fixture(scope='package')
def fix_error_rpc():
	@axon.worker.rpc()
	def throw_error():
		raise BaseException('Calling this RPC will raise an error')

@pytest.mark.tl
@pytest.mark.asyncio
async def test_error_catching(fix_error_rpc):
	url = f'{url_scheme}://localhost:{axon.config.transport.config.port}/rpc'
	ss = axon.client.get_stub(url)

	with pytest.raises(BaseException) as err:
		await ss.throw_error()

	assert str(err.value) == 'Calling this RPC will raise an error'

@pytest.mark.tl
def test_tl_basic(fix_live_transport_worker):
	tlw, port = fix_live_transport_worker
	
	tlc = TransportClient()

	def wrk_fn(param):
		return param

	tlw.register_RPC(wrk_fn, '/test_tl_basic/wrk_fn', axon.inline_executor.InlineExecutor())

	url = f'{url_scheme}://localhost:{port}/test_tl_basic/wrk_fn'
	result = tlc.call_rpc(url, ('hello!', ), {})
	assert(result == 'hello!')

@pytest.mark.tl
@pytest.mark.asyncio
async def test_second_tl():

	port = axon.utils.get_open_port(lower_bound=8002)
	tlc = TransportClient()
	tlw = TransportWorker(port)

	class TestClass():

		def __init__(self):
			pass

		def test_fn(self, param):
			return param

	t = TestClass()
	axon.worker.service(t, 'test_second_tl_service',tl=tlw)

	worker_thread = threading.Thread(target=tlw.run, daemon=True)
	worker_thread.start()

	time.sleep(0.5)

	# the positive test that the service registered with the secondary transport layer exists
	url = f'{url_scheme}://localhost:{port}/test_second_tl_service'
	ss = axon.client.get_stub(url, tl=tlc)
	result = await ss.test_fn('hi there!')
	assert(result == 'hi there!')

	# the negative test that the service registered with the secondary transport layer does not show up on the primary transport layer
	rw = axon.client.get_stub('localhost')
	assert(hasattr(rw, 'test_second_tl_service') == False)




import axon
import asyncio
import time
import threading
import multiprocessing
import inspect
import pytest

from itertools import product
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

url_scheme = axon.config.url_scheme
message = 'hello world!'

start_delay = 2*0.05
small_delay = 2*0.05
big_delay = 2*0.5

class PoolTestService():
	async def print_this_async(self, delay, message):
		loop = asyncio.get_event_loop()
		await asyncio.sleep(delay)
		return message

	def print_this(self, delay, message):
		time.sleep(delay)
		return message

@pytest.fixture(scope='package')
def fix_pool_services():
	ils = PoolTestService()
	axon.worker.service(ils, name='inline_service')

	tps = PoolTestService()
	tpe = ThreadPoolExecutor(max_workers=10)
	axon.worker.service(tps, name='thread_pool_service', executor=tpe)

	tps = PoolTestService()
	ppe = ProcessPoolExecutor(max_workers=10, mp_context=multiprocessing.get_context("spawn"))
	axon.worker.service(tps, name='process_pool_service', executor=ppe)

@pytest.mark.tl
@pytest.mark.asyncio
async def test_invokation(fix_pool_services):
	remote_worker = axon.client.get_stub(f'{url_scheme}://localhost:{axon.config.transport.config.port}')
	delay = 0

	assert(await remote_worker.inline_service.print_this(delay, message) == message)
	assert(await remote_worker.thread_pool_service.print_this(delay, message) == message)
	assert(await remote_worker.process_pool_service.print_this(delay, message) == message)

	assert(await remote_worker.inline_service.print_this_async(delay, message) == message)
	assert(await remote_worker.thread_pool_service.print_this_async(delay, message) == message)
	assert(await remote_worker.process_pool_service.print_this_async(delay, message) == message)

	print('test_invokation complete!')

@pytest.mark.tl
@pytest.mark.asyncio
async def test_inline_concurrency(fix_pool_services):
	# tests that non async services with the inline executor do not run in parallel

	inline_service = axon.client.get_stub(f'localhost:/inline_service')
	completion_order = []

	async def long_coro():
		await inline_service.print_this(big_delay, 'inline')
		completion_order.append('long_coro')

	async def short_coro():
		time.sleep(start_delay)
		await inline_service.print_this(small_delay, 'inline')
		completion_order.append('short_coro')

	reqs = [long_coro(), short_coro()]
	await asyncio.gather(*reqs)

	assert(completion_order[0] == 'long_coro')
	assert(completion_order[1] == 'short_coro')

async def verify_parallel(service):
	# this function tests if calls to a service can be run in parallel by making two calls, a long call, and a short call within that one. If the short call finishes first despite starting last, the test passes

	completion_order = []

	async def long_coro():
		await service(big_delay, message)
		completion_order.append('long_coro')

	async def short_coro():
		time.sleep(start_delay)
		await service(small_delay, message)
		completion_order.append('short_coro')

	reqs = [long_coro(), short_coro()]
	await asyncio.gather(*reqs)

	assert(completion_order[0] == 'short_coro')
	assert(completion_order[1] == 'long_coro')

@pytest.mark.tl
@pytest.mark.asyncio
async def test_service_concurrency(fix_pool_services):
	# this function tests to see if calls to services run in executors can be run in parallel
	
	endpoints = list(product(['thread_pool_service', 'process_pool_service'], ['print_this', 'print_this_async']))
	# the async version of print_this should run in parallel, even with the inline executor
	endpoints.append(('inline_service', 'print_this_async'))
	services = [axon.client.get_stub(f'localhost/{name}/{v}') for name, v in endpoints]

	for s in services:
		print(s.url)
		await verify_parallel(s)



from axon.transport_client import AsyncResultHandle, req_executor

from urllib.parse import urlparse, urlunparse
import asyncio

def add_url_defaults(url, config):
	
	if '://' not in url:
		url = config.scheme+'://'+url

	comps = urlparse(url)

	if (comps.port == None):
		netloc = f'{comps.hostname}:{config.port}'
		url = urlunparse((comps.scheme, netloc, comps.path, '', '' , ''))

	return url

class GenericStub():

	def __init__(self, tl, url):
		self.url = add_url_defaults(url, tl.get_config())
		self.tl = tl

	def __call__(self, *args, **kwargs):
		return AsyncResultHandle(self.tl.call_rpc, self.url, args, kwargs)

class FutureStub(GenericStub):

	def __init__(self, tl, url):
		GenericStub.__init__(self, tl, url)

	def __call__(self, *args, **kwargs):
		return req_executor.submit(self.tl.call_rpc, self.url, args, kwargs)

class SyncStub(GenericStub):

	def __init__(self, tl, url):
		GenericStub.__init__(self, tl, url)

	def __call__(self, *args, **kwargs):
		return self.tl.call_rpc(self.url, args, kwargs)


import sys
sys.path.append('..')

from flask import Flask
from flask import request as route_req
from concurrent.futures import ProcessPoolExecutor as PPE

import traceback
import logging

from axon.transport_worker import AbstractTransportWorker, invoke_RPC
from axon.serializers import serialize
from axon.HTTP_transport import config

class HTTPTransportWorker(AbstractTransportWorker):

	def __init__(self, port=config.port):
		# all RPCs registered with this TL are stored here in this dict
		self.rpcs = {}

		# removes the startup text
		cli = sys.modules['flask.cli']
		cli.show_server_banner = lambda *x: None
		log = logging.getLogger('werkzeug')
		log.disabled = True

		# the app that listens for incomming http requests
		self.app = Flask('HTTPTransportWorker')	
		self.port = port

		# the route that listens for incomming RPC requests
		@self.app.route('/', defaults={'path': ''}, methods=['POST'])
		@self.app.route('/<path:path>', methods=['POST'])
		def catch_all(path):
			
			try:
				path = '/'+path
				(fn, executor) = self.rpcs[path]

				param_str = route_req.form['msg']
				result_str = executor.submit(invoke_RPC, fn, param_str, in_parallel=True).result()
				result_str = f'0|{result_str}'

			except:
				result_str = serialize((traceback.format_exc(), sys.exc_info()[1]))
				result_str = f'1|{result_str}'

			return result_str

	def run(self):
		self.app.run(host='0.0.0.0', port=self.port)


import sys
sys.path.append('..')

import axon
import cloudpickle

itlw = axon.reflector.ITLW(url='143.198.32.69', name='host_worker')

@axon.worker.rpc(tl=itlw)
# @axon.worker.rpc()
def host(service_str, name):

	service = cloudpickle.loads(service_str)
	axon.worker.register_ServiceNode(service, name)
	itlw.update_profile()
	return 'success'

axon.worker.init()


import axon
import time
import pytest
import threading
import pickle

default_depth = 3
url_scheme = axon.config.url_scheme
TransportWorker = type(axon.config.default_service_config['tl'])

class DummyClass():

	def __init__(self, depth=default_depth):
		self.child = None
		self.depth = depth

		if (depth>0):
			self.child = DummyClass(depth=depth-1)

	def test_fn(self):
		print('test fn called at depth: '+str(self.depth))
		pass

def test_add_child_not_live():

	t = DummyClass()
	s = axon.worker.ServiceNode(t, 'test')

	def test_add_child_rpc(self):
		print('test_add_child_rpc called')

	s.add_child('test_add_child', test_add_child_rpc)

	# checks that the parent node has test_add_child_rpc as a child
	assert('test_add_child' in s.children)

	# checks that the endpoint prefix is correct
	child_config = s.children['test_add_child'].children['__call__']
	assert(child_config['endpoint_prefix'] == '/test/test_add_child')

	# checks that the ServiceNode's profile contains the added child
	p = s.get_profile()
	assert('test_add_child' in p)

	print('test_add_child_child_config passed!')

def test_remove_child_not_live():

	t = DummyClass()
	s = axon.worker.ServiceNode(t, 'test')

	s.remove_child('test_fn')

	# checks that the ServiceNode no longer has a child called test_fn
	assert('test_fn' not in s.children)

	# checks that the ServiceNode's profile does not contain the added child
	p = s.get_profile()
	assert('test' not in p)

	# checks that the necessary RPCs have been removed from the transport layer
	assert('/test/test_fn' not in s.tl.rpcs)
	assert('/test/test_fn/__call__' not in s.tl.rpcs)

@pytest.fixture(scope='package')
def fix_live_transport_worker():
	port = axon.utils.get_open_port(lower_bound=8003)
	tlw = TransportWorker(port)

	worker_thread = threading.Thread(target=tlw.run, daemon=True)
	worker_thread.start()
	time.sleep(0.5)

	return tlw, port

def test_add_child_live(fix_live_transport_worker):
	tlw, port = fix_live_transport_worker

	t = DummyClass()
	parent = axon.worker.ServiceNode(t, 'test_add_child_live', tl=tlw)

	child = DummyClass()

	parent.add_child('test_add_child', child)

	stub = axon.client.get_stub(f'{url_scheme}://localhost:{port}/test_add_child_live')

	stub.test_add_child.test_fn().join()

def test_remove_child_live(fix_live_transport_worker):
	tlw, port = fix_live_transport_worker

	t = DummyClass()
	parent = axon.worker.ServiceNode(t, 'test_remove_child_live', tl=tlw)

	# the stub before the child gets removed
	stub_with_child = axon.client.get_stub(f'{url_scheme}://localhost:{port}/test_remove_child_live')

	parent.remove_child('child')

	# recursively checks that all the child's children were removed as well
	for i in range(default_depth-1):
		print(i)
		stub_with_child = stub_with_child.child

		with pytest.raises(KeyError):
			stub_with_child.test_fn().join()

	# the stub after the child gets removed
	no_child_stub = axon.client.get_stub(f'{url_scheme}://localhost:{port}/test_remove_child_live')

	# checks that the child attribute has been removed from the stub
	assert(not hasattr(no_child_stub, 'child'))


from .config import comms_config

import socket
import copy
import random

def get_ID_generator(n=10_000):
	L = list(range(n))
	random.shuffle(L)
	while True:
		for l in L:
			yield str(l)

def get_active_workers():
	status, text = GET('http://'+str(comms_config.notice_board_ip)+':'+str(comms_config.notice_board_port)+'/get_ips')

	if (status == 404): print('notice board not found')

	return deserialize(text)

self_ip = None
def get_self_ip():
	global self_ip

	if (self_ip == None):
		s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
		s.connect(("8.8.8.8", 80))
		ip = s.getsockname()[0]
		s.close()

		self_ip = ip

		return ip

	else:
		return self_ip

# scans for an available port in between upper_bound and lower_bound
def get_open_port(lower_bound=8000, upper_bound=9000):

	for target_port in range(lower_bound, upper_bound):
		# defines a socket, AF_INIT means ipv4 and SOCK_STREAM means TCP
		sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		location = ("127.0.0.1", target_port)
		# will return 0 if the port is available
		result = sock.connect_ex(location)

		if result != 0:
			sock.close()
			return target_port

	sock.close()
	raise BaseException('No available ports in between '+str(lower_bound)+' and '+str(upper_bound))

# accepts two dicts, target and source
# in any shared keys between the two will be overwritten to source's value, and any keys in source will be copied to target, with thei values
def overwrite(target, source):
	# to avoid aliasing
	target = copy.copy(target)

	for key in source:
		target[key] = source[key]

	return target


# the bit that exports worker profile class and uses the RVL
# an RPC stub is the thing on the client that makes a calling request and waits for the response

from axon.config import default_service_config, default_rpc_endpoint, default_client_tl
from axon.stubs import GenericStub, add_url_defaults

from types import SimpleNamespace

transport_client = default_client_tl

def get_stub(url, tl=transport_client, stub_type=GenericStub, top_stub_type=object):
	url = add_url_defaults(url, tl.get_config())
	profile = tl.call_rpc(url, (), {})
	url_components = url.split('/')
	base_url = '/'.join(url_components[:3])
	return make_ServiceStub(base_url, tl, profile, stub_type, top_stub_type)

def make_ServiceStub(url, tl, profile, stub_type=GenericStub, top_stub_type=object):

	attrs = {}
	keys = list(profile.keys())
	parent_classes = (top_stub_type, )
	banned_keys = ['__call__', '__profile_flag__', '__func__', '__self__', '__get__', '__set__', '__delete__'] + dir(object())

	for key in keys:

		if (key in banned_keys): continue

		member = profile[key]

		if '__profile_flag__' in member:
			# member is a profile for a ServiceNode
			attrs[key] = make_ServiceStub(url, tl, member, stub_type, object)

		else:
			# If a member is not a profile, then it must be an RPC config, and so correspond to a callable object on the worker with no __dict__attribute
			stub_url = url + f"{member['endpoint_prefix']}/{key}"
			attrs[key] = stub_type(tl=tl, url=stub_url)

	if '__call__' in keys:
		# if the profile has a __call__ attribute, than the corresponding object on the server is callable and has a __dict__ attribute, and so must be represented by an RPC stub bound to the given network coordinates
		stub_url = url + f"{profile['__call__']['endpoint_prefix']}/__call__"
		BoundStubClass = get_BoundStubClass(stub_type, tl, stub_url)
		# this ensures the stub will inherit from a stub class that's bound to the configuration
		parent_classes = (BoundStubClass, ) + parent_classes

	ServiceStub = type('ServiceStub', parent_classes, attrs)
	return ServiceStub()

def get_BoundStubClass(stub_type, tl, url):
	
	# a class for stubs that are bound to a certain RPC
	class BoundStubClass(stub_type):
		def __init__(self):
			stub_type.__init__(self, tl, url)

	return BoundStubClass


import psutil
import time
import logging
import os

if (__name__ == '__main__'):
	logger = logging.getLogger(__name__)
	logger.setLevel(logging.INFO)
	handler = logging.FileHandler('./reflector.metrics')
	log_format = logging.Formatter('%(asctime)s | %(message)s')
	handler.setFormatter(log_format)
	logger.addHandler(handler)

	time.sleep(1)
	cpu = psutil.cpu_percent()
	ram = psutil.virtual_memory().percent
	net = psutil.net_io_counters()

	logger.info('CPU %s RAM %s bytes_sent %s bytes_recv %s', cpu, ram, net.bytes_sent, net.bytes_recv)



import pathlib
from setuptools import setup, find_packages
# from axon.config import version

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
	name="axon-ECRG",
	version="0.2.4",
	description="RPC proxy framework",
	long_description=README,
	long_description_content_type="text/markdown",
	url="https://github.com/DuncanMays/axon-ECRG",
	packages=['axon', 'axon.tests', 'axon.HTTP_transport', 'axon.reflector', 'axon.reflector.tests'],
	package_data={'': ['pytest.ini']},
	install_requires=[
		'pytest',
		'pytest-asyncio',
		'flask',
		'urllib3',
		'requests', 
		'aiohttp', 
		'python-socketio[client]', 
		'psutil',
		'cloudpickle'
	],
)




from axon.utils import overwrite
from axon.config import default_service_config, default_service_depth

from copy import copy

transport_layers = set()

class ServiceNode():

	def __init__(self, subject, name, depth=default_service_depth, **configuration):
		self.subject = subject
		self.name = name
		self.children = {}
		self.depth = depth

		self.configuration = overwrite(default_service_config, configuration)
		self.tl = self.configuration['tl']

		# iterates over members and either registers them as RPCs or recursively turns them into ServiceNodes
		for key in dir(self.subject):

			# dir calls the overwritable __dir__ function on self.subject. This means it's unreliable in some cases, since developers can specify what it returns.
			# this can mean the attribute is listed but does not exist, or even that calling the attribute throws an error
			try:
				if not hasattr(self.subject, key):
					continue

			except(BaseException):
				# sometimes merely accessing an attribute can throw an error
				continue

			member = getattr(self.subject, key)

			if (key == '__call__'):
				# Any member accessed via __call__ attribute is represented in profile by an RPC config. This means that objects stored on __call__ attributes will not be represented in profile
				self.init_RPC(key, member)
			
			elif hasattr(member, '__dict__'):
				# Any member with a __dict__ attribute gets a profile, provided it as not been accessed via __call__
				self.add_child(key, member)

			elif hasattr(member, '__call__'):
				# Any member with a __call__ attribute but no __dict__ attribute is represented in profile by an RPC config
				self.init_RPC(key, member)

		# we now register an RPC at the ServiceNode's endpoint to expose its profile
		profile_endpoint = f"{self.configuration['endpoint_prefix']}/{self.name}"
		self.tl.register_RPC(self.get_profile, profile_endpoint, default_service_config['executor'])
		transport_layers.add(self.tl)

	def remove_child(self, child_key):

		child = self.children[child_key]

		if isinstance(child, ServiceNode):
			# remove all grandchildren through the child to be deleted by calling that child's remove_child method
			grand_child_keys = list(child.children.keys())
			for grand_child_key in grand_child_keys:

				if (grand_child_key == '__call__'):
					endpoint = f"{self.configuration['endpoint_prefix']}/{self.name}/{child_key}/{grand_child_key}"
					self.tl.deregister_RPC(endpoint)

				else:
					child.remove_child(grand_child_key)

		# deregisters the child's profile endpoint
		profile_endpoint = f"{self.configuration['endpoint_prefix']}/{self.name}/{child_key}"
		self.tl.deregister_RPC(profile_endpoint)

		del self.children[child_key]

	def add_child(self, key, child, **child_config):

		# limits recursion to a depth parameter
		if (self.depth < 0): return

		# The child config overwrites the parent's config, meaning by default children inherit configuration from their parents
		child_config = overwrite(self.configuration, child_config)

		if self.name != '':
			# the child's endpoint_prefix is the parent's prefix plus / and the child's name
			child_config['endpoint_prefix'] += '/'+str(self.name)

		# create a ServiceNode out of it and register it as a child
		child = ServiceNode(child, key, depth=self.depth-1, **child_config)
		self.children[key] = child
		return child

	def init_RPC(self, key, fn):

		tl = self.configuration['tl']
		endpoint = f"{self.configuration['endpoint_prefix']}/{self.name}/{key}"

		# this dict will be sent back with the profile to client
		child_config = {
			'endpoint_prefix': self.configuration['endpoint_prefix'] +'/'+ str(self.name),
			'name': key,
			'executor': self.configuration['executor']
		}

		tl.register_RPC(fn, endpoint, self.configuration['executor'])

		# remember the configuration
		self.children[key] = child_config

	# returns a JSON serializable dict tree with leaves of RPC configuration dicts
	def get_profile(self):
		
		profile = {}

		# this is a flag attribute that lets a ServiceStub know that a dict is a profile for a ServiceNode, not a configuration dict for an RPC
		profile['__profile_flag__'] = True

		for key in self.children.keys():
			child = self.children[key]

			# if the child is itself a ServiceNode, recursively get the profiles of its attributes
			if isinstance(child, ServiceNode):
				
				# if the child uses a different transport layer, ignore it
				if (self.tl != child.tl):
					continue

				profile[key] = child.get_profile()

			# if the attribute is an RPC, the profile element is its configuration
			else:
				child_copy = copy(child)
				del child_copy['executor']
				profile[key] = child_copy

		return profile


import sys
sys.path.append('..')

from axon.transport_worker import invoke_RPC
from axon.serializers import serialize, deserialize
from axon.chunking import send_in_chunks, recv_chunks
from axon.stubs import add_url_defaults
from axon.worker import TLSNs

from concurrent.futures import ProcessPoolExecutor as PPE
from types import SimpleNamespace
from math import ceil

import socketio
import cloudpickle
import time
import sys
import traceback

class ITLW():

	def __init__(self, name, url='http://143.198.32.69:5000'):
		self.name = name
		self.reflector_url = add_url_defaults(url, SimpleNamespace(port=5000, scheme='http'))
		self.rpcs = {}
		self.sio = None

	def run(self):

		self.sio = socketio.Client()
		self.sio.connect(self.reflector_url)

		# this is the first message we'll send the reflector, containing the worker name
		self.sio.emit('worker_header', data=str(self.name))

		self.update_profile()

		@self.sio.event
		def rpc_request(req_str):
			call_ID, endpoint, param_str = req_str.split('|', 3)

			return_object = {
				'errcode': 0,
				'result': None,
			}

			try:
				(fn, executor) = self.rpcs[endpoint]

				result_str = executor.submit(invoke_RPC, fn, param_str).result()
				result_str = f'0|{result_str}'

			except:
				result_str = serialize((traceback.format_exc(), sys.exc_info()[1]))
				result_str = f'1|{result_str}'

			chunk_size = 100_000

			if (len(result_str) < chunk_size):
				self.sio.emit('rpc_result', data=f'{call_ID}|{result_str}')

			else:
				num_chunks = ceil(len(result_str)/chunk_size)

				for i in range(num_chunks):
					chunk_str = result_str[ chunk_size*i : chunk_size*(i+1) ]
					self.sio.emit('rpc_result_chunk', data=f'{str(i)}|{str(num_chunks)}|{call_ID}|{chunk_str}')

		while True:
			time.sleep(1_000_000)

	def update_profile(self):
		
		tlsn = TLSNs[id(self)]
		profile = tlsn.get_profile()
		self.sio.emit('update_profile', data=serialize(profile))

	def register_RPC(self, fn, endpoint, executor):

		if isinstance(executor, PPE):
			fn = cloudpickle.dumps(fn)
		
		self.rpcs[endpoint] = (fn, executor)

	def deregister_RPC(self, endpoint):

		if endpoint in self.rpcs:
			del self.rpcs[endpoint]
		else:
			raise BaseException(f'No RPC registered at endpoint: {endpoint}')